# Autogenerated by Astropy's setup.py on 2019-12-17 05:34:12 UTC
import datetime

version = "4.0"
githash = "763f069058cb1ee2038fa1cfa485452ec94b726e"


major = 4
minor = 0
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 12, 17, 5, 34, 12)
debug = False

astropy_helpers_version = "4.0"
